"""
contains the business logic (namely the BDEW calendar information)

"""
from datetime import date

from holidays import HolidayBase, HolidaySum
from holidays.constants import DEC
from holidays.countries.germany import Germany


class BdewDefinedHolidays(HolidayBase):
    """
    Special holidays considered by the BDEW, excluding nationwide or state specific holidays.

    To get a full calendar with all holidays use the function `create_bdew_calendar`.

    Currently, in Germany there are no observed holidays e.g. a Holiday is
    moved to the next Workday, if it's on Sunday.

    """

    def __init__(self, observed: bool = False, **kwargs):
        super().__init__(observed=observed, **kwargs)

    def _populate(self, year):
        self[date(year, DEC, 24)] = "Heiligabend"
        self[date(year, DEC, 31)] = "Silvester"


def create_bdew_calendar() -> HolidaySum:
    """Creates a calendar containing all days considered by the BDEW.

    Currently, in Germany there are no observed holidays e.g. a Holiday is
    moved to the next Workday, if it's on Sunday.

    "Nach GPKE und GeLi Gas sind alle jene Tage Werktage, die weder Sonnabend,
    Sonntag oder gesetzlicher Feiertag sind. Generell gelten der 24.12. und
    der 31.12. als Feiertage. Angaben ohne Gewähr."

    https://www.bdew.de/media/documents/GPKE-GeLiGas-Feiertagskalender-2022-Version-1-1_Uu1w6UQ.pdf

    [Accessed 2022-11-02]``

    Returns
    -------
    holidays.HolidaySum
        dict-like holiday calendar, to check if a date/datetime is a holiday
    """

    # First we need the BDEW specific holidays.
    calendar = BdewDefinedHolidays()

    # the type is wrong at assignment but correct after the first loop interation
    result: HolidaySum = calendar  # type:ignore[assignment]

    # If a day is holiday in any subdivision, the holiday is valid nationwide.
    # Therefore, we add all subdivisions of Germany to the BDEW specific holidays.
    # Currently, in Germany holidays are not observed.
    for subdivision in Germany.subdivisions:
        # the method __add__ expects a Union[int, "HolidayBase", "HolidaySum"] as `other`
        # here, we're dealing with a child instance of HolidayBase
        result += Germany(
            subdiv=subdivision, observed=False
        )  # type:ignore[assignment]

    return result



"""
periods is a module that helps to calculate statutory periods
("gesetzliche Fristen") in the German energy market.
It is based on the chapter "Fristenberechnung" in the GPKE.
"""
import datetime
import sys
from dataclasses import dataclass
from datetime import date
from enum import Enum
from typing import Optional, Union

from dateutil.relativedelta import relativedelta
from holidays import SAT, SUN

from bdew_datetimes import GERMAN_TIME_ZONE, create_bdew_calendar

try:
    from typing import Literal
except ImportError:  # for Python 3.7
    pass

# https://www.bundesnetzagentur.de/DE/Beschlusskammern/1_GZ/BK6-GZ/2020/BK6-20-160/Mitteilung_Nr_2/Leseversion_GPKE.pdf
# pages 15 onwards

_bdew_calendar = create_bdew_calendar()
"""
a static calendar object that is used (module) internally
"""


class DayType(str, Enum):
    """
    An enum to differentiate between calendar days and working days.
    """

    WORKING_DAY = "WT"  #: working day, German "Werktag"
    CALENDAR_DAY = "KT"  #: calendar day, German "Kalendertag"


class EndDateType(Enum):
    """
    An enum to distinguish inclusive and exclusive end dates.
    """

    INCLUSIVE = 1
    """
    If a contract ends with the year 2022 and the end date is denoted as "2022-12-31",
    then the end date is inclusive. Most dates in human (spoken) communication are meant
    inclusively.
    """

    EXCLUSIVE = 2
    """
    If a constract ends with the year 2022 and the end date is denoted as "2023-01-01",
    then the end date is exclusive. Most end dates handled by technical systems are meant
    exclusively.
    """


if sys.version_info.minor > 7:
    _DayTyp = Union[DayType, Literal["WT", "KT"]]
else:
    _DayTyp = Union[DayType, str]  # type:ignore[misc]


@dataclass
class Period:
    """
    A period is a German "Frist": A tuple that consists of a number of days and a day type.
    """

    number_of_days: int
    """
    number of days (might be any value <0, >0 or ==0)
    """
    day_type: DayType
    """
    the kind of days to add/subtract
    """

    def __init__(
        self,
        number_of_days: int,
        day_type: _DayTyp,
        end_date_type: EndDateType = EndDateType.EXCLUSIVE,
    ):
        """
        Initialize the Period by providing a number of days and a day_type which define the period.

        """
        self.number_of_days = number_of_days
        # If the Period is about something ending (e.g. a contract), then the user may
        # provide an end_date_type.
        # Internally we handle all end dates as exclusive, because:
        # https://hf-kklein.github.io/exclusive_end_dates.github.io/
        if end_date_type == EndDateType.INCLUSIVE:
            if self.number_of_days > 0:
                self.number_of_days = self.number_of_days - 1
            elif self.number_of_days < 0:
                self.number_of_days = self.number_of_days + 1
        if isinstance(day_type, DayType):
            pass
        elif isinstance(day_type, str):
            day_type = DayType(day_type)
        else:
            raise ValueError(
                f"'{day_type}' is not an allowed value; Check the typing"
            )
        self.day_type: DayType = day_type


def is_bdew_working_day(candidate: date) -> bool:
    """
    Returns true if and only if the given candidate is a day relevant for the period calculation.
    Returns false if the given candidate is either a BDEW holiday, a saturday or sunday.
    """
    if candidate in _bdew_calendar:
        return False
    return candidate.weekday() not in (SAT, SUN)


def get_next_working_day(start_date: date) -> date:
    """
    If start_date is a working day, the next (+1) working day is returned.
    If this day is a BDEW holiday or falls on a weekend, the next working day
    is returned.
    """
    result = start_date + datetime.timedelta(
        days=1
    )  # in any case the calculation starts at least at the next day
    while not is_bdew_working_day(result):
        # if the next day is a holiday, then we add days until we find a working day
        result += datetime.timedelta(days=1)
    return result


def get_previous_working_day(start_date: date) -> date:
    """
    If start_date is a working day, the previous (-1) working day is returned.
    If this day is a BDEW holiday or falls on a weekend, the previous working day
    is returned.
    """
    result = start_date - datetime.timedelta(
        days=1
    )  # in any case the calculation starts at least at the next day
    while not is_bdew_working_day(result):
        # if the previous day is a holiday, then we subtract days until we find a working day
        result -= datetime.timedelta(days=1)
    return result


def add_frist(start: date, period: Period) -> date:
    """
    Returns the date that is period after start.
    """
    result: date = start
    if period.number_of_days >= 0:
        # the period calculation starts at the next working day, even if the number_of_days == 0
        result = get_next_working_day(result)
    # result is now the "Beginndatum" of the Fristenberechnung
    if period.day_type == DayType.CALENDAR_DAY:
        return result + datetime.timedelta(days=period.number_of_days)
    # day_type is working day
    if period.number_of_days == 0:
        return result
    if period.number_of_days > 0:
        days_added = 0
        while days_added < period.number_of_days:
            result += datetime.timedelta(days=1)
            if is_bdew_working_day(result):
                # it's a usual working day; those are considered for the period calculation
                days_added += 1
    else:  # implicitly: elif period.number_of_days < 0:
        days_subtracted = 0
        while days_subtracted <= abs(period.number_of_days):
            result -= datetime.timedelta(days=1)
            if is_bdew_working_day(result):
                days_subtracted += 1
    return result


class MonthType(Enum):
    """
    When calculating periods defined as 'nth working day of a month' the
    BNetzA regulations distinguish between two types of month which are
    modelled in this enum.
    Some periods refer to the "Liefermonat", others to the "Fristenmonat".
    """

    LIEFERMONAT = 1
    """
    The "Liefermonat" is the month in which the supply starts.
    """
    FRISTENMONAT = 2
    """
    The grid operators prefer a key date based handling of supply contracts.
    The key date in these cases is usually expressed as a specific working day
    in the so called "Fristenmonat".
    The "Fristenmonat" starts at the first day of the month
    _before_ the "Liefermonat".
    Quote: 'Nach der Festlegung BK6-06-009 (GPKE) der Monat vor dem Liefermonat.'
    """
    # pylint:disable=line-too-long
    # source: https://www.bundesnetzagentur.de/DE/Beschlusskammern/1_GZ/BK6-GZ/_bis_2010/2006/BK6-06-009/BK6-06-009_Beschluss_download.pdf?__blob=publicationFile&v=5


def get_nth_working_day_of_month(
    number_of_working_day_in_month: int,
    month_type: MonthType = MonthType.LIEFERMONAT,
    start: Optional[date] = None,
) -> date:
    """
    Returns the nth working of the month, starting at start.
    If start is None, then we use the local date/start of month in Germany.
    Only year and month of the provided start date are taken into account.
    The date.day is always discarded.

    This function is useful if you're dealing with statutory periods from the
    GPKE/MaBiS that are defined as "working days since start of month".
    """
    if start is None:
        start = GERMAN_TIME_ZONE.localize(datetime.datetime.utcnow()).date()
    if month_type == MonthType.LIEFERMONAT:
        # returns the "nter Werktag des Liefermonats"
        start = get_previous_working_day(start.replace(day=1))
        period = Period(
            number_of_days=number_of_working_day_in_month,
            day_type=DayType.WORKING_DAY,
            end_date_type=EndDateType.INCLUSIVE,
        )
        result = add_frist(start, period)
    elif month_type == MonthType.FRISTENMONAT:
        # returns the "nter Werktag des Fristenmonats"
        start_of_next_month = start.replace(day=1) + relativedelta(months=1)
        result = get_nth_working_day_of_month(
            number_of_working_day_in_month,
            month_type=MonthType.LIEFERMONAT,
            start=start_of_next_month,
        )
    else:
        raise ValueError(f"Unhandled month_type {month_type}")
    return result



from datetime import datetime, timedelta, timezone

import pytest  # type:ignore[import]

from bdew_datetimes.german_strom_and_gas_tag import (
    GERMAN_TIME_ZONE,
    Division,
    has_no_utc_offset,
    is_gastag_limit,
    is_stromtag_limit,
    is_xtag_limit,
)


@pytest.mark.parametrize(
    "dt, expected_is_start_or_end_of_german_stromtag",
    [
        pytest.param(
            datetime(2020, 1, 1, 0, 0, 0, tzinfo=timezone.utc), False
        ),
        pytest.param(
            datetime(2019, 12, 31, 23, 0, 0, tzinfo=timezone.utc), True
        ),
        pytest.param(
            datetime(2019, 12, 31, 22, 0, 0, tzinfo=timezone.utc), False
        ),
        pytest.param(
            datetime(2010, 1, 1, 0, 0, 0, tzinfo=GERMAN_TIME_ZONE), True
        ),
        pytest.param(
            datetime(
                2010, 1, 1, 0, 0, 0, tzinfo=timezone(timedelta(seconds=3600))
            ),
            True,
        ),
        pytest.param(
            datetime(2022, 3, 26, 23, 0, 0, 0, tzinfo=timezone.utc), True
        ),
        pytest.param(
            datetime.fromisoformat("2022-03-27T00:00:00+01:00"), True
        ),
        pytest.param(
            datetime.fromisoformat("2022-03-27T00:00:00+02:00"), False
        ),
        pytest.param(
            datetime(2022, 3, 27, 22, 0, 0, 0, tzinfo=timezone.utc), True
        ),
        pytest.param(
            datetime.fromisoformat("2022-03-28T00:00:00+02:00"), True
        ),
        pytest.param(
            datetime(2022, 10, 29, 22, 0, 0, 0, tzinfo=timezone.utc), True
        ),
        pytest.param(
            datetime.fromisoformat("2022-10-30T00:00:00+02:00"), True
        ),
        pytest.param(
            datetime(2022, 10, 30, 23, 0, 0, 0, tzinfo=timezone.utc), True
        ),
        pytest.param(
            datetime.fromisoformat("2022-10-31T00:00:00+01:00"), True
        ),
        pytest.param(
            datetime.fromisoformat("2022-10-31T00:00:00+02:00"), False
        ),
        pytest.param(
            datetime.fromisoformat("2022-10-29T12:00:00-10:00"),
            True,
            id="Hawaii, German DST",
        ),
        pytest.param(
            datetime.fromisoformat("2022-10-30T13:00:00-10:00"),
            True,
            id="Hawaii, German standard time",
        ),
        pytest.param(
            datetime.fromisoformat("2022-10-30T07:00:00+09:00"),
            True,
            id="Tokyo, German DST",
        ),
        pytest.param(
            datetime.fromisoformat("2022-10-31T08:00:00+09:00"),
            True,
            id="Tokyo, German standard time",
        ),
    ],
)
def test_stromtag(
    dt: datetime, expected_is_start_or_end_of_german_stromtag: bool
):
    actual = is_stromtag_limit(dt)
    assert actual == expected_is_start_or_end_of_german_stromtag


@pytest.mark.parametrize(
    "dt, expected_is_start_or_end_of_german_gastag",
    [
        pytest.param(
            datetime(2020, 1, 1, 0, 0, 0, tzinfo=timezone.utc), False
        ),
        pytest.param(datetime(2020, 1, 1, 5, 0, 0, tzinfo=timezone.utc), True),
        pytest.param(
            datetime(2020, 1, 1, 4, 0, 0, tzinfo=timezone.utc), False
        ),
        pytest.param(
            datetime(2022, 3, 26, 5, 0, 0, 0, tzinfo=timezone.utc), True
        ),
        pytest.param(
            datetime.fromisoformat("2022-03-26T06:00:00+01:00"), True
        ),
        pytest.param(
            datetime.fromisoformat("2022-03-27T04:00:00+02:00"), False
        ),
        pytest.param(
            datetime(2022, 3, 27, 4, 0, 0, 0, tzinfo=timezone.utc), True
        ),
        pytest.param(
            datetime.fromisoformat("2022-03-27T06:00:00+02:00"), True
        ),
        pytest.param(
            datetime(2022, 10, 29, 4, 0, 0, 0, tzinfo=timezone.utc), True
        ),
        pytest.param(
            datetime.fromisoformat("2022-10-30T06:00:00+01:00"), True
        ),
        pytest.param(
            datetime.fromisoformat("2022-10-30T06:00:00+02:00"), False
        ),
        pytest.param(
            datetime.fromisoformat("2022-10-29T19:00:00-10:00"),
            True,
            id="Hawaii, German DST",
        ),
        pytest.param(
            datetime.fromisoformat("2022-10-30T19:00:00-10:00"),
            True,
            id="Hawaii, German standard time",
        ),
        pytest.param(
            datetime.fromisoformat("2022-10-29T09:45:00+05:45"),
            True,
            id="Nepal, German DST",
        ),
        pytest.param(
            datetime.fromisoformat("2022-10-30T10:45:00+05:45"),
            True,
            id="Nepal, German standard time 1",
        ),
        pytest.param(
            datetime.fromisoformat("2022-10-31T10:45:00+05:45"),
            True,
            id="Nepal, German standard time 2",
        ),
    ],
)
def test_gastag(dt: datetime, expected_is_start_or_end_of_german_gastag: bool):
    actual = is_gastag_limit(dt)
    assert actual == expected_is_start_or_end_of_german_gastag


@pytest.mark.parametrize(
    "dt, expected_has_not_utc_offset",
    [
        pytest.param(datetime(2020, 1, 1, 0, 0, 0, tzinfo=timezone.utc), True),
        pytest.param(
            datetime.fromisoformat("2022-03-26T06:00:00+01:00"), False
        ),
        pytest.param(
            datetime.fromisoformat("2022-03-27T04:00:00+02:00"), False
        ),
    ],
)
def test_has_no_utc_offset(dt: datetime, expected_has_not_utc_offset: bool):
    actual = has_no_utc_offset(dt)
    assert actual == expected_has_not_utc_offset


@pytest.mark.parametrize(
    "dt, division",
    [
        pytest.param(datetime(2020, 1, 1, 0, 0, 0, tzinfo=timezone.utc), 0),
    ],
)
def test_is_xtag_limit_raises(dt: datetime, division: Division):
    with pytest.raises(NotImplementedError):
        is_xtag_limit(date_time=dt, division=division)


@pytest.mark.parametrize(
    "dt, division, expected",
    [
        pytest.param(
            datetime(2020, 1, 1, 0, 0, 0, tzinfo=timezone.utc),
            Division.GAS,
            False,
        ),
        pytest.param(
            datetime(2020, 1, 1, 5, 0, 0, tzinfo=timezone.utc),
            Division.GAS,
            True,
        ),
        pytest.param(
            datetime(2020, 1, 1, 4, 0, 0, tzinfo=timezone.utc),
            Division.GAS,
            False,
        ),
        pytest.param(
            datetime(2022, 3, 26, 5, 0, 0, 0, tzinfo=timezone.utc),
            Division.GAS,
            True,
        ),
        pytest.param(
            datetime(2020, 1, 1, 0, 0, 0, tzinfo=timezone.utc),
            Division.STROM,
            False,
        ),
        pytest.param(
            datetime(2019, 12, 31, 23, 0, 0, tzinfo=timezone.utc),
            Division.STROM,
            True,
        ),
        pytest.param(
            datetime(2019, 12, 31, 22, 0, 0, tzinfo=timezone.utc),
            Division.STROM,
            False,
        ),
        pytest.param(
            datetime(2010, 1, 1, 0, 0, 0, tzinfo=GERMAN_TIME_ZONE),
            Division.STROM,
            True,
        ),
    ],
)
def test_is_xtag_limit(dt: datetime, division: Division, expected):
    actual = is_xtag_limit(dt, division)
    assert actual == expected



"""
A module to evaluate datetimes and whether they are "on the edge"
of a German "Stromtag" or "Gastag" respectively
"""
from datetime import datetime, time
from enum import Enum
from typing import Callable

# The problem with the stdlib zoneinfo is, that the availability of timezones
# via ZoneInfo(zone_key) depends on the OS and system on which you're running
# it. In some cases "Europe/Berlin" might be available, but generally it's not,
# and it's PITA to manually define timezones. So we're using pytz as a
# datasource for timezone information.
from pytz import utc

from bdew_datetimes import GERMAN_TIME_ZONE


class Division(Enum):
    """
    Allows to distinguish divisions used by German utilities, German "Sparte".
    """

    STROM = 1  #: electricity
    GAS = 2  #: gas


def _get_german_local_time(date_time: datetime) -> time:
    """
    Returns the current german local time for the given datetime object.
    """
    german_local_datetime = date_time.astimezone(GERMAN_TIME_ZONE)
    return german_local_datetime.time()


def has_no_utc_offset(date_time: datetime) -> bool:
    """
    Returns true if and only if date_time has an explicit offset to UTC.
    If the UTC-offset is 0, it is exactly "+00:00".
    """
    # the name of the function contains a negation because in German
    # market communication it often matters that the UTC offset is 0.
    original_time = date_time.time()
    utc_time = date_time.astimezone(tz=utc).time()
    return (
        utc_time == original_time
        and utc_time.hour == 0
        and utc_time.minute == 0
        and utc_time.second == 0
    )


def is_stromtag_limit(
    date_time: datetime,
) -> bool:  # the name is not as speaking as I'd like it to be
    """
    Returns true if and only if the given date_time is the inclusive
    start or exclusive end of a german "Stromtag".
    The "Stromtag" is the balancing relevant day in the German electricity
    market.
    It starts and ends at midnight in German local time which can be
    either 23:00 h or 22:00 h in UTC (depending on the daylight saving time in Germany).
    """
    german_local_time = _get_german_local_time(date_time)
    return (
        german_local_time.hour == 0
        and german_local_time.minute == 0
        and german_local_time.second == 0
    )


def is_gastag_limit(
    date_time: datetime,
) -> bool:  # the name is not as speaking as I'd like it to be
    """
    Returns true if and only if the given date_time is the inclusive start
    or exclusive end of a german "Gastag".
    The "Gastag" is the balancing relevant day in the German gas market.
    It starts and ends at 6am in German local time which can be either
    04:00 h or 05:00 h in UTC (depending on the daylight saving time in Germany).
    """
    german_local_time = _get_german_local_time(date_time)
    return (
        german_local_time.hour == 6
        and german_local_time.minute == 0
        and german_local_time.second == 0
    )


def is_xtag_limit(date_time: datetime, division: Division) -> bool:
    """
    Evaluates if it is the start/end of a provided division.
    """
    xtag_evaluator: Callable[[datetime], bool]
    if division == Division.STROM:
        xtag_evaluator = is_stromtag_limit
    elif division == Division.GAS:
        xtag_evaluator = is_gastag_limit
    else:
        raise NotImplementedError(
            f"The division must either be 'Strom' or 'Gas': '{division}'"
        )
    return xtag_evaluator(date_time)



from datetime import date, datetime

import pytest  # type:ignore[import]
from holidays import DateLike

from bdew_datetimes.calendar import BdewDefinedHolidays, create_bdew_calendar


@pytest.mark.parametrize(
    "expected_holiday",
    [
        pytest.param(date(2021, 12, 31), id="Silvester 2021"),
        pytest.param(date(2021, 12, 24), id="Heiligabend 2021"),
        pytest.param(date(2070, 12, 31), id="arbitrary Silvester"),
        pytest.param(date(2070, 12, 24), id="arbitrary Heiligabend"),
    ],
)
def test_bdew_holidays(expected_holiday: date):
    calendar = BdewDefinedHolidays()
    assert expected_holiday in calendar


@pytest.mark.parametrize(
    "test_date, expected_is_in_calendar",
    [
        pytest.param(date(2021, 12, 31), True, id="Silvester 2021 (BDEW)"),
        pytest.param(date(2022, 1, 1), True, id="Neujahr 2022 (bundesweit)"),
        pytest.param(date(2022, 1, 2), False, id="a regular Sunday"),
        pytest.param(date(2022, 8, 15), True, id="Mariä Himmelfahrt (BY, SL)"),
    ],
)
def test_create_bdew_calendar(test_date: date, expected_is_in_calendar: bool):
    calendar = create_bdew_calendar()
    if expected_is_in_calendar:
        assert test_date in calendar
    else:
        assert test_date not in calendar


def test_holiday_calendar_obj():
    calendar = create_bdew_calendar()

    assert not calendar.observed
    assert calendar.country == "DE"


@pytest.mark.parametrize(
    "candidate",
    [
        pytest.param(datetime(2022, 1, 1, 22, 16, 59)),
        pytest.param("2022-01-01"),
        pytest.param("1/1/2022"),
        pytest.param(
            datetime(2022, 1, 1, 22, 16, 59).timestamp(),
            id="POSIX timestamp: 1641071819.0",
        ),
        pytest.param(int(datetime(2022, 1, 1, 22, 16, 59).timestamp())),
    ],
)
def test_holiday_in_calendar(candidate: DateLike):
    calendar = create_bdew_calendar()
    assert candidate in calendar



from datetime import date

import pytest  # type:ignore[import]

from bdew_datetimes.periods import (
    DayType,
    EndDateType,
    MonthType,
    Period,
    _DayTyp,
    add_frist,
    get_next_working_day,
    get_nth_working_day_of_month,
    get_previous_working_day,
)


def test_period_instantiation_with_enum():
    period = Period(42, DayType.WORKING_DAY)
    assert period.number_of_days == 42
    assert period.day_type == DayType.WORKING_DAY


def test_period_instantiation_with_enum_and_inclusive_end():
    period = Period(
        42, DayType.WORKING_DAY, end_date_type=EndDateType.INCLUSIVE
    )
    assert period.number_of_days == 41
    assert period.day_type == DayType.WORKING_DAY


@pytest.mark.parametrize(
    "number_of_days, str_daytype, expected",
    [
        pytest.param(42, "WT", Period(42, DayType.WORKING_DAY)),
        pytest.param(17, "KT", Period(17, DayType.CALENDAR_DAY)),
    ],
)
def test_period_instantiation_with_str(
    number_of_days: int, str_daytype: _DayTyp, expected: Period
):
    actual = Period(number_of_days, str_daytype)
    assert actual == expected


def test_instantiation_with_invalid_str():
    with pytest.raises(ValueError):
        _ = Period(
            42, "Foo"  # type:ignore[argument] # literal is intentionally wrong
        )


@pytest.mark.parametrize(
    "start,expected",
    [
        pytest.param(
            date(2023, 1, 1),
            date(2023, 1, 2),
        ),
        pytest.param(
            date(2023, 1, 2), date(2023, 1, 3), id="regular next working day"
        ),
        pytest.param(
            date(2023, 1, 5),
            date(2023, 1, 9),
            id="Skip Hl.drei Könige (subdivision holiday) and weekend",
        ),
        pytest.param(
            date(2022, 12, 30),
            date(2023, 1, 2),
            id="Skip Silvester (BDEW) and Neujahr (federal) holidays",
        ),
    ],
)
def test_get_next_working_day(start: date, expected: date):
    actual = get_next_working_day(start)
    assert actual == expected


@pytest.mark.parametrize(
    "start,expected",
    [
        pytest.param(
            date(2023, 1, 1),
            date(2022, 12, 30),
        ),
        pytest.param(
            date(2023, 1, 3),
            date(2023, 1, 2),
            id="regular previous working day",
        ),
        pytest.param(
            date(2023, 1, 9),
            date(2023, 1, 5),
            id="Skip Hl.drei Könige (subdivision holiday) and weekend",
        ),
    ],
)
def test_get_previous_working_day(start: date, expected: date):
    actual = get_previous_working_day(start)
    assert actual == expected


@pytest.mark.parametrize(
    "start,frist,expected",
    [
        pytest.param(
            date(2023, 1, 1),
            Period(0, DayType.CALENDAR_DAY),
            date(2023, 1, 2),
            id="+0 KT is the next day if start is a holiday",
        ),
        pytest.param(
            date(2023, 1, 2),
            Period(0, DayType.CALENDAR_DAY),
            date(2023, 1, 3),
            id="+0 KT is the next working day if start is a working day",
        ),
        pytest.param(
            date(2023, 1, 1),
            Period(0, DayType.WORKING_DAY),
            date(2023, 1, 2),
            id="+0 WT days is next (working) day if start is a holiday",
        ),
        pytest.param(
            date(2023, 1, 1),
            Period(1, DayType.CALENDAR_DAY),
            date(2023, 1, 3),
            id="+1 KT is day after next working day if start is a holiday",
        ),
        pytest.param(
            date(2023, 1, 3),
            Period(-1, DayType.CALENDAR_DAY),
            date(2023, 1, 2),
            id="-1 KT is day before working day if start is a holiday",
        ),
        pytest.param(
            date(2023, 1, 2),
            Period(1, DayType.CALENDAR_DAY),
            date(2023, 1, 4),
            id="+1 KT is day after next working day if start is a working day",
        ),
        pytest.param(
            date(2023, 1, 1),
            Period(1, DayType.WORKING_DAY),
            date(2023, 1, 3),
            id="BDEW Holiday + 1 working day is start",
        ),
        pytest.param(
            # From the Fristenberechnung-Chapter in the GPKE document:
            # "Lieferende bei Lieferantenwechselvorgängen:
            # Eingang der Abmeldung des LFA erfolgt am 04.07.2016. Der Mindestzeitraum von sechs WT beginnt am
            # 05.07.2016 und endet am 12.07.2016. Frühestes zulässiges Abmeldedatum ist damit der 12.07.2016 [...]."
            date(2016, 7, 4),
            Period(
                6,
                DayType.WORKING_DAY,
                end_date_type=EndDateType.INCLUSIVE
                # the abmeldedatum is the inclusive end of a contract
            ),
            date(2016, 7, 12),
            id="Lieferende bei Lieferantenwechselvorgängen",
        ),
        pytest.param(
            date(2016, 7, 12),
            Period(
                -6,
                DayType.WORKING_DAY,
                end_date_type=EndDateType.INCLUSIVE
                # the abmeldedatum is the inclusive end of a contract
            ),
            date(2016, 7, 4),
            id="revert 'Lieferende bei Lieferantenwechselvorgängen'",
        ),
        pytest.param(
            # Eingang der Anmeldung des LFN erfolgt am 04.07.2016. Der Mindestzeitraum von sieben bzw. zehn WT
            # beginnt am 05.07.2016 und endet am 13.07.2016 bzw. 18.07.2016. Frühestes zulässiges Anmeldedatum
            # ist damit der 14.07.2016 bzw. 19.07.2016, sodass die Marktlokation dem LFN frühestens zum Beginn
            # des vorgenannten Tages zugeordnet wird.
            date(2016, 7, 4),
            Period(
                7,
                DayType.WORKING_DAY,
                end_date_type=EndDateType.EXCLUSIVE
                # lieferbeginn is the exclusive end of the previous contract
            ),
            date(2016, 7, 14),
            id="Lieferbeginn bei Lieferantenwechselvorgängen (7WT)",
        ),
        pytest.param(
            date(2016, 7, 14),
            Period(
                -7,
                DayType.WORKING_DAY,
                end_date_type=EndDateType.EXCLUSIVE
                # lieferbeginn is the exclusive end of the previous contract
            ),
            date(2016, 7, 4),
            id="revert 'Lieferbeginn bei Lieferantenwechselvorgängen' (-7WT)",
        ),
        pytest.param(
            # Eingang der Anmeldung des LFN erfolgt am 04.07.2016. Der Mindestzeitraum von sieben bzw. zehn WT
            # beginnt am 05.07.2016 und endet am 13.07.2016 bzw. 18.07.2016. Frühestes zulässiges Anmeldedatum
            # ist damit der 14.07.2016 bzw. 19.07.2016, sodass die Marktlokation dem LFN frühestens zum Beginn
            # des vorgenannten Tages zugeordnet wird.
            date(2016, 7, 4),
            Period(
                10,
                DayType.WORKING_DAY,
                end_date_type=EndDateType.EXCLUSIVE
                # lieferbeginn is the exclusive end of the previous contract
            ),
            date(2016, 7, 19),
            id="Lieferbeginn bei Lieferantenwechselvorgängen (10WT)",
        ),
        pytest.param(
            date(2016, 7, 19),
            Period(
                -10,
                DayType.WORKING_DAY,
                end_date_type=EndDateType.EXCLUSIVE
                # lieferbeginn is the exclusive end of the previous contract
            ),
            date(2016, 7, 4),
            id="revert 'Lieferbeginn bei Lieferantenwechselvorgängen' (-10WT)",
        ),
    ],
)
def test_add_frist(start: date, frist: Period, expected: date):
    actual = add_frist(start, frist)
    assert actual == expected


@pytest.mark.parametrize(
    "number,start,month_type,expected",
    [
        pytest.param(
            14,
            date(2023, 1, 1),
            MonthType.LIEFERMONAT,
            date(2023, 1, 20),
            id="14ter Werktag des (Liefer)Monats Januar 203; BK-Summen",
        ),
        pytest.param(
            16,
            date(2023, 1, 1),
            MonthType.LIEFERMONAT,
            date(2023, 1, 24),
            id="16ter Werktag des (Liefer)Monats Januar 2023; Zuordnungslisten",
        ),
        pytest.param(
            17,
            date(2023, 1, 1),
            MonthType.LIEFERMONAT,
            date(2023, 1, 25),
            id="17ter Werktag des (Liefer)Monats Januar 2023; BK-Zuordnungsliste, Deklarationsliste",
        ),
        pytest.param(
            18,
            date(2023, 11, 1),
            MonthType.LIEFERMONAT,
            date(2023, 11, 28),
            id="18ter Werktag des (Liefer)Monats November 2023; Deklarationsmitteilung",
        ),
        pytest.param(
            14,
            date(2023, 2, 1),
            MonthType.LIEFERMONAT,
            date(2023, 2, 20),
            id="14ter Werktag des (Liefer)Monats Februar 2023; BK-Summen",
        ),
        pytest.param(
            16,
            date(2023, 2, 1),
            MonthType.LIEFERMONAT,
            date(2023, 2, 22),
            id="16ter Werktag des (Liefer)Monats Februar 2023; Zuordnungslisten",
        ),
        pytest.param(
            17,
            date(2023, 2, 1),
            MonthType.LIEFERMONAT,
            date(2023, 2, 23),
            id="17ter Werktag des (Liefer)Monats Februar 2023; BK-Zuordnungsliste, Deklarationsliste",
        ),
        pytest.param(
            18,
            date(2023, 2, 1),
            MonthType.LIEFERMONAT,
            date(2023, 2, 24),
            id="18ter Werktag des (Liefer)Monats Februar 2023; Deklarationsmitteilung",
        ),
        pytest.param(
            21,
            date(2023, 1, 1),
            MonthType.FRISTENMONAT,
            date(2023, 3, 1),
            id="21ter Werktag des Fristenmonats Januar 2023;",
        ),
        pytest.param(
            21,
            date(2023, 2, 1),
            MonthType.FRISTENMONAT,
            date(2023, 3, 30),
            id="21ter Werktag des Fristenmonats Februar; NKP",
        ),
        pytest.param(
            26,
            date(2022, 12, 1),
            MonthType.FRISTENMONAT,
            date(2023, 2, 7),
            id="26ter Werktag des Fristenmonats Dezember 2022",
        ),
        pytest.param(
            26,
            date(2023, 1, 1),
            MonthType.FRISTENMONAT,
            date(2023, 3, 9),
            id="26ter Werktag des Fristenmonats Januar 2023",
        ),
        pytest.param(
            26,
            date(2023, 6, 1),
            MonthType.FRISTENMONAT,
            date(2023, 8, 7),
            id="26ter Werktag des Fristenmonats Juni 2023; NKP MG-Überlappung",
        ),
        pytest.param(
            30,
            date(2023, 4, 1),
            MonthType.FRISTENMONAT,
            date(2023, 6, 15),
            id="30ter Werktag des Fristenmonats April 2023; letztmalige Datenabnahme beim BIKO",
        ),
        pytest.param(
            42,
            date(2022, 11, 1),
            MonthType.FRISTENMONAT,
            date(2023, 1, 31),
            id="42ter Werktag des Fristenmonats November 2022; BK-Abrechnung",
        ),
        pytest.param(
            42,
            date(2023, 7, 1),
            MonthType.FRISTENMONAT,
            date(2023, 9, 29),
            id="42ter Werktag des Fristenmonats Juli 2023; BK-Abrechnung",
        ),
    ],
)
def test_get_nth_working_day_of_month(
    number: int, start: date, month_type: MonthType, expected: date
):
    actual = get_nth_working_day_of_month(number, month_type, start)
    assert actual == expected



"""
bdew_datetimes is a package that models the BDEW holiday, which is relevant for German utilities
"""
from pytz import timezone

from .calendar import create_bdew_calendar

GERMAN_TIME_ZONE = timezone("Europe/Berlin")

__version__ = "0.2.0"
